#include <cuda_runtime.h>
#include <cstdio>

#include "scheduler.h"

using namespace std;

namespace OptChol
{

extern __shared__ double dynamic_shared[];

__device__ bool getThreadInfo(const int & threadsperstid, const int & problemsperblock, unsigned int * const & jobs,const int & jobssize, int & localidx, unsigned int & threadstid, unsigned int & threadstnum, unsigned int & blockproblemnum)
{
	int threadnum = blockIdx.x*blockDim.x + threadIdx.x;
	threadstnum = threadnum/threadsperstid;
	blockproblemnum = threadstnum%problemsperblock;
	localidx = threadnum%threadsperstid;
	if (threadstnum >= jobssize) return false;

	threadstid = jobs[threadstnum];

	return true;
}

__device__ bool isAllInactive(bool active)
{
	__shared__ int reduce[16];
	int thidx = threadIdx.y*blockDim.x + threadIdx.x;
	int pred = __all(active);
	if (thidx<16)
	{
		reduce[thidx]=1;
	}
	__syncthreads();
	if (thidx&31==0)
	{
		reduce[thidx>>5]=pred;
	}
	__syncthreads();
	if (thidx<16)
	{
		pred=__all(reduce[thidx]);
	}
	__syncthreads();
	if (thidx==0)
	{
		reduce[0]=pred;
	}
	__syncthreads();

	return reduce[0]==0;
}

__device__ void m_copy_nonzeros(int * shared_nonzeros, int * nonzeros, void *, const int & idx)
{
	shared_nonzeros[idx] = nonzeros[idx];
}

__device__ void m_copy_fortranidx(int * shared_idx, int * fortranidx, void *, const int & idx)
{
	shared_idx[idx] = fortranidx[idx]-1;
}

__device__ void m_copy_invsdiag(int * invsdiag, int * sdiag, void *, const int & idx)
{
	invsdiag[sdiag[idx]-1] = idx;
}

__device__ void m_copy_to_shared(double * dest, double * src, int * map, const int & idx)
{
	dest[idx] = src[map[idx]];
}

__device__ void m_simple_copy(double * dest, double * src, void *, const int & idx)
{
	dest[idx] = src[idx];
}

__device__ void m_simple_int_copy(int * dest, int * src, void *, const int & idx)
{
	dest[idx] = src[idx];
}

__device__ void m_copy_to_device(double * dest, double * src, int * map, const int & idx)
{
//	printf("copy col[%d] => A[%d] = %f \n",  idx, map[idx], src[idx]);
	dest[map[idx]] = src[idx];
}

__device__ void m_div(double * dest, int memsize, double val, const int & idx)
{
	dest[idx*memsize] /= val;
}

__device__ void m_diagdiv(double * A, int * mapped_rhselim, double diag, const int & idx)
{
//	printf("col[%d] /= %f\n", mapped_rhselim[idx], diag);
	A[mapped_rhselim[idx]] /= diag;
}

__device__ void m_matrixelim(int colmemsizeA, double * A, double * col, int * nonzeros, int * mapped_matrixelim, const int & idx)
{
	int i = mapped_matrixelim[2*idx];
	int j = mapped_matrixelim[2*idx+1];
	A[nonzeros[j]*colmemsizeA + nonzeros[i]] -= col[i]*col[j];
//	printf("A[%d,%d] -= col[%d]*col[%d]\n", nonzeros[i], nonzeros[j], i,j);
}


__device__ void m_rhselim(double * colB, double * colA, int * rhselim, int diagidx, void*, const int & idx)
{
	int i = rhselim[idx]-1;
	colB[i] -= colB[diagidx]*colA[i];
}

/*
__device__ void m_rhselim(double * colB, double * colA, int * mapped_rhselim, int diagidx, void*, const int & idx)
{
	int i = mapped_rhselim[idx];
	colB[i] -= colB[diagidx]*colA[i];
}
*/
__device__ void m_rhselimprint(double * colB, double * colA, int * mapped_rhselim, int diagidx, void*, const int & idx)
{
	int i = mapped_rhselim[idx];
	colB[i] -= colB[diagidx]*colA[i];
	printf("elim=%d, val=%e, %e, %e\n", i, colB[i], colB[diagidx], colA[i]);
}


__device__ void m_backsubst(double * colB, double * colA, int * backsubst, int diagidx, void*, const int & idx)
{
	int i = backsubst[idx]-1;
	colB[i] -= colB[diagidx]*colA[i];
}

/*
__device__ void m_backsubst(double * colB, double * colA, double B_entry, const int & idx)
{
	colB[idx] -= B_entry*colA[idx];
}
*/
template <typename T1, typename T2, typename T3, typename T4, typename T5>
__device__ void exec_arg5(void(*func)(T1, T2, T3, T4, T5, const int & ), const int & size, const int & localidx, const int & threadsperstid, int & copied_counter, const bool & active,
	const T1 & arg1, const T2 & arg2, const T3 & arg3, const T4 & arg4, const T5 & arg5)
{
	int _copied_counter = 0;

	while (_copied_counter < size)
	{
		int copycount = min(size-_copied_counter, threadsperstid);
		if (localidx < copycount && active)
		{
			func(arg1, arg2, arg3, arg4, arg5, _copied_counter+localidx);
		}

		_copied_counter += copycount;
	}
}

template <typename T1, typename T2, typename T3>
__device__ void exec_arg3(void(*func)(T1, T2, T3, const int &), const int & size, const int & localidx, const int & threadsperstid, int & copied_counter, const bool & active,
	const T1 & arg1, const T2 & arg2, const T3 & arg3)
{
	int _copied_counter = 0;

	while (_copied_counter < size)
	{
		int copycount = min(size-_copied_counter, threadsperstid);
		if (localidx < copycount && active)
		{
			func(arg1, arg2, arg3, _copied_counter+localidx);
		}
		_copied_counter += copycount;
	}
}

__global__ void kernel_chlsco(Scheduler::T_CUDADATA data, int totalmemsizeA, int colmemsizeA, double * A, int * returncode, unsigned int *  jobs,  int jobssize)
{
	double * col;
	int * shared_nonzeros;
	int * copied_counter;
	double * stidA;
	int * error;

	__shared__ int * nonzeros;
	__shared__ int * mapped_rhselim;
	__shared__ int * mapped_matrixelim;
	__shared__ int cnt_nonzeros;
	__shared__ int cnt_rhselim;
	__shared__ int cnt_matrixelim;
	__shared__ int i1;
	__shared__ int mapped_i1;


	int localidx=0;
	unsigned int threadstid=0;
	unsigned int threadstnum=0;
	unsigned int blockproblemnum=0;

	bool active = getThreadInfo(data.threadsperstid, data.problemsperblock, jobs, jobssize, localidx, threadstid, threadstnum, blockproblemnum);
	active = active && returncode[threadstid]==0;

	if (isAllInactive(active)) return;


/*	if (active)
	{
		printf("data. threadsperstid=%d, jobssize=%d\n", data.threadsperstid, jobssize);
		printf("stid=%d, localidx=%d, threadstnum=%d, blockproblemnum=%d, data.problemsperblock=%d\n", threadstid, localidx, threadstnum, blockproblemnum, data.problemsperblock);
	}
*/
	col = dynamic_shared+blockproblemnum*data.n;
	shared_nonzeros = (int*)(dynamic_shared+data.problemsperblock*data.n);
	copied_counter = shared_nonzeros+data.n;
	error = copied_counter+data.problemsperblock;

	stidA = A+threadstid*totalmemsizeA;

	__syncthreads();
	if (threadIdx.x==0)
	{
		nonzeros = data.nonzeros;
		mapped_rhselim = data.mapped_rhselim;
		mapped_matrixelim = data.mapped_matrixelim;
	}

	if (localidx == 0 && active)
	{
		error[blockproblemnum] = false;
	}

	__syncthreads();

	for (int i=0; i<data.n-1; i++)
	{
		if (threadIdx.x == 0)
		{
			cnt_nonzeros = data.n_nonzeros[i];
			cnt_rhselim = data.n_rhselim[i];
			cnt_matrixelim = data.n_matrixelim[i];
			i1 = data.sdiag[i]-1;
			mapped_i1 = data.mapped_sdiag[i];
/*			printf("i1=%d\n", i1);
			for (int j=0; j<39; j++)
			{
				printf("A[%d*%d+%d]=%f\n", i1,colmemsizeA,j, stidA[i1*colmemsizeA+j]);
			}
*/		}
//	printf("error=%p\n", error);

		__syncthreads();

		// Nonzeros Shared Memory kopieren
		exec_arg3(
			m_copy_nonzeros, cnt_nonzeros, threadIdx.x, blockDim.x, copied_counter[0], 1, 
			shared_nonzeros, nonzeros, (void*)0);

		__syncthreads();

		// Spalte in Shared Memory kopieren
		exec_arg3(
			m_copy_to_shared, cnt_nonzeros, localidx, data.threadsperstid, copied_counter[blockproblemnum], active, 
			col, stidA+i1*colmemsizeA, shared_nonzeros);

		__syncthreads();

		if (localidx==0 && active)
		{
			double tdiag = col[mapped_i1];
			if (tdiag <= 0.0)
			{
				error[blockproblemnum]=true;
			} else
			{
				tdiag = sqrt(tdiag);
				col[mapped_i1] = tdiag;
//				printf("col[%d]=diag=%f\n", mapped_i1, diag[blockproblemnum]);
			}
		}

		__syncthreads();

		if (active && error[blockproblemnum]) active=false;

		// rhs durch diag teilen
		exec_arg3(
			m_diagdiv, cnt_rhselim, localidx, data.threadsperstid, copied_counter[blockproblemnum], active, 
			col, mapped_rhselim, col[mapped_i1]);

		__syncthreads();

		if(threadIdx.x==0) mapped_rhselim += cnt_rhselim;

		__syncthreads();

		// Spalte zurück in Device Memory kopieren
		exec_arg3(
			m_copy_to_device, cnt_nonzeros, localidx, data.threadsperstid, copied_counter[blockproblemnum], active,
			stidA+i1*colmemsizeA, col, shared_nonzeros);

		__syncthreads();

		exec_arg5(
			m_matrixelim, cnt_matrixelim, localidx, data.threadsperstid, copied_counter[blockproblemnum], active, 
			colmemsizeA, stidA, col, shared_nonzeros, mapped_matrixelim);

		__syncthreads();

		if(threadIdx.x==0)
		{
			mapped_matrixelim += cnt_matrixelim*2;
			nonzeros += cnt_nonzeros;
		}
	}

	if (localidx == 0 && active)
	{
		int i1 = data.sdiag[data.n-1]-1;
		double diag = stidA[i1*colmemsizeA+i1];
		if (diag <= 0.0)
		{
			error[blockproblemnum]=true;
		} else
		{
			diag = sqrt(diag);
			stidA[i1*colmemsizeA+i1] = diag;
		}
	}

	if (localidx == 0 && error[blockproblemnum])
	{
		returncode[threadstid] = 1;
	}

}

__global__ void kernel_transp(Scheduler::T_CUDADATA data, int totalmemsizeA, int colmemsizeA, double * A, int * returncode, unsigned int *  jobs,  int jobssize)
{
	bool active=true;
	int threadnumy = blockIdx.x*blockDim.y + threadIdx.y;
	int problemsperblock = blockDim.y/blockDim.x;
	int threadstnum = threadnumy/blockDim.x;
	int sharedmemshift = (blockDim.x&1==0) ? (blockDim.x+1) : blockDim.x;

	if (threadstnum >= jobssize) active=false;

	int localidxy = threadnumy%blockDim.x;

	int threadstid = jobs[threadstnum];
	int blockproblemnum = threadIdx.y/blockDim.x;

	active = active && returncode[threadstid]==0;

	if (isAllInactive(active)) return;

	double * stidA = A+threadstid*totalmemsizeA;
	double * buffer = dynamic_shared+blockproblemnum*blockDim.x*blockDim.x;

	__shared__ int * invsdiag;
	__shared__ int copied_counter;
	if (threadIdx.x == 0 && threadIdx.y==0)
	{
		invsdiag = (int*)(dynamic_shared+problemsperblock*blockDim.x*blockDim.x);
	}

	int subblocks = data.n/blockDim.x;
	if (data.n%blockDim.x != 0) subblocks++;

//	printf("stid=%d, localidxy=%d, threadstnum=%d, blockproblemnum=%d, blockDim.x=%d\n", threadstid, localidxy, threadstnum, blockproblemnum, blockDim.x);

	__syncthreads();

	exec_arg3(
		m_copy_invsdiag, data.n, blockDim.x*threadIdx.y + threadIdx.x, problemsperblock*blockDim.x*blockDim.x, copied_counter, 1, 
		invsdiag, data.sdiag, (void*)0);

	__syncthreads();

	int row;
	int col;

	for (int rowsub=0; rowsub < subblocks; rowsub ++)
	{
		for (int colsub=0; colsub <= rowsub; colsub++)
		{
//			if (threadIdx.x==0 && threadIdx.y==0) printf("read\n");
			if (active)
			{
				col = colsub*blockDim.x + localidxy;
				row = rowsub*blockDim.x + threadIdx.x;

				// Teilmatrix einlesen
				if (row < data.n && col < row && invsdiag[row] > invsdiag[col] )
				{
//					printf("stid %d: buffer(%d, %d) <= A(%d,%d)\n", threadstid, threadIdx.x, localidxy, row, col);
					buffer[localidxy*sharedmemshift + threadIdx.x] = stidA[colmemsizeA*col + row];
				}

			}
			__syncthreads();

//			if (threadIdx.x==0 && threadIdx.y==0) printf("readwrite\n");

			if (active)
			{
				col = rowsub*blockDim.x + localidxy;
				row = colsub*blockDim.x + threadIdx.x;

				// Teilmatrix transponiert schreiben
				if (col < data.n && row < col)
				{
					if (invsdiag[row] < invsdiag[col])
					{
//						printf("stid %d: A(%d, %d) <= buffer(%d,%d)\n", threadstid, row, col, localidxy, threadIdx.x);
						stidA[colmemsizeA*col + row] = buffer[threadIdx.x*sharedmemshift + localidxy];
					}
					else
					{
//						printf("stid %d: TRANS buffer(%d, %d) <= A(%d,%d)\n", threadstid, localidxy, threadIdx.x, row, col);
						buffer[threadIdx.x*sharedmemshift + localidxy] = stidA[colmemsizeA*col + row];
					}
				}

			}
			__syncthreads();
//			if (threadIdx.x==0 && threadIdx.y==0) printf("write\n");

			if (active)
			{
				col = colsub*blockDim.x + localidxy;
				row = rowsub*blockDim.x + threadIdx.x;

				// Teilmatrix einlesen
				if (row < data.n && col < row && invsdiag[row] < invsdiag[col])
				{
//					printf("stid %d: TRANS A(%d,%d) <=  buffer(%d, %d) \n", threadstid, row, col, threadIdx.x, localidxy);
					stidA[colmemsizeA*col + row] = buffer[localidxy*sharedmemshift + threadIdx.x];
				}

			}
			__syncthreads();

		}
	}

}

/*
__global__ void kernel_rrslvo(Scheduler::T_CUDADATA data,
	int totalmemsizeA, int colmemsizeA, double * A,
	int m, int totalmemsizeB, int colmemsizeB, double * B,
	int * returncode, unsigned int *  jobs,  int jobssize)
{


	double * colA;
	double * colB;
	int * shared_nonzeros;
	int * shared_backsubst;
	int * copied_counter;
	double * stidA;
	double * stidB;

	__shared__ int * nonzeros;
	__shared__ int * backsubst;
	__shared__ int * mapped_rhselim;
	__shared__ int cnt_nonzeros;
	__shared__ int cnt_rhselim;
	__shared__ int cnt_backsubst;
	__shared__ int i1;
	__shared__ int mapped_i1;

	int localidx = 0;
	unsigned int threadstid = 0;
	unsigned int threadstnum = 0;
	unsigned int blockproblemnum = 0;

	bool active = getThreadInfo(data.rhs_threadsperstid, data.rhs_problemsperblock, jobs, jobssize, localidx, threadstid, threadstnum, blockproblemnum);
	active = active && returncode[threadstid]==0;

	if (isAllInactive(active)) return;

//	printf("stid=%d, localidx=%d, threadstnum=%d, blockproblemnum=%d, data.problemsperblock=%d\n", threadstid, localidx, threadstnum, blockproblemnum, data.rhs_problemsperblock);
//	if (threadIdx.x==0)	printf("===================================================\n\nRRSLV\n\n===================================================\n");

	colA = dynamic_shared+blockproblemnum*data.n;
	colB = colA+data.rhs_problemsperblock*data.n;
	shared_backsubst = shared_nonzeros = (int*)(dynamic_shared+2*data.rhs_problemsperblock*data.n);
	copied_counter = shared_nonzeros+data.n;

	stidA = A+threadstid*totalmemsizeA;
	stidB = B+threadstid*totalmemsizeB;

	__syncthreads();
	if (threadIdx.x==0)
	{
		nonzeros = data.nonzeros;
		mapped_rhselim = data.mapped_rhselim;
	}

	__syncthreads();

	for (int i=0; i<data.n-1; i++)
	{
		if (threadIdx.x == 0)
		{
			cnt_nonzeros = data.n_nonzeros[i];
			cnt_rhselim = data.n_rhselim[i];
			i1 = data.sdiag[i]-1;
			mapped_i1 = data.mapped_sdiag[i];
		}
//	printf("error=%p\n", error);

		__syncthreads();

		// Nonzeros Shared Memory kopieren
		exec_arg3(
			m_copy_nonzeros, cnt_nonzeros, threadIdx.x, blockDim.x, copied_counter[0], 1, 
			shared_nonzeros, nonzeros, (void*)0);

		__syncthreads();

		// Spalte in Shared Memory kopieren
		exec_arg3(
			m_copy_to_shared, cnt_nonzeros, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
			colA, stidA+i1*colmemsizeA, shared_nonzeros);

		__syncthreads();

		for (int k=0; k<m; k++)
		{
			exec_arg3(
				m_copy_to_shared, cnt_nonzeros, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
				colB, stidB+k*colmemsizeB, shared_nonzeros);

			__syncthreads();

			if (localidx==0 && active)
			{
				colB[mapped_i1] /= colA[mapped_i1];
			}

			__syncthreads();

			exec_arg5(
				m_rhselim, cnt_rhselim, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
				colB, colA, mapped_rhselim, mapped_i1, (void*)0);

			__syncthreads();

			// Spalte zurück in Device Memory kopieren
			exec_arg3(
				m_copy_to_device, cnt_nonzeros, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active,
				stidB+k*colmemsizeB, colB, shared_nonzeros);

			__syncthreads();

		}

		__syncthreads();
		if(threadIdx.x==0)
		{
			mapped_rhselim += cnt_rhselim;
			nonzeros += cnt_nonzeros;
		}
	}

	if (threadIdx.x==0)
	{
		backsubst = data.backsubst;
		cnt_backsubst = data.n_backsubst[0];
		i1 = data.sdiag[data.n-1]-1;
	}
	__syncthreads();

	if (localidx == 0 && active)
	{
		double diag = stidA[i1*colmemsizeA+i1];
		for (int k=0; k<m; k++)
		{
			stidB[k*colmemsizeB+i1] /= (diag*diag);
		}
	}

	__syncthreads();

	for (int i=data.n-1; i>=1; i--)
	{
		// Nonzeros Shared Memory kopieren
		exec_arg3(
			m_copy_fortranidx, cnt_backsubst, threadIdx.x, blockDim.x, copied_counter[0], 1, 
			shared_backsubst, backsubst, (void*)0);

		__syncthreads();

		// Spalte in Shared Memory kopieren
		exec_arg3(
			m_copy_to_shared, cnt_backsubst, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
			colA, stidA+i1*colmemsizeA, shared_backsubst);

		__syncthreads();

		for (int k=0; k<m; k++)
		{
			exec_arg3(
				m_copy_to_shared, cnt_backsubst, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
				colB, stidB+k*colmemsizeB, shared_backsubst);

			__syncthreads();

			exec_arg3(
				m_backsubst, cnt_backsubst, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
				colB, colA, stidB[k*colmemsizeB+i1]);

			__syncthreads();

			// Spalte zurück in Device Memory kopieren
			exec_arg3(
				m_copy_to_device, cnt_backsubst, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active,
				stidB+k*colmemsizeB, colB, shared_backsubst);

			__syncthreads();
		}
		if (threadIdx.x == 0)
		{
			backsubst += cnt_backsubst;
			cnt_backsubst = data.n_backsubst[data.n-(i-1)-1];
			i1 = data.sdiag[i-1]-1;
		}

		__syncthreads();

		double diag = stidA[i1*colmemsizeA+i1];

		if (localidx==0 && active)
		{
			for (int k=0; k<m; k++)
			{
				stidB[k*colmemsizeB+i1] /= diag;
			}

		}

		__syncthreads();

	}

}
*/


__global__ void kernel_rrslvo(Scheduler::T_CUDADATA data,
	int totalmemsizeA, int colmemsizeA, double * A,
	int m, int totalmemsizeB, int colmemsizeB, double * B,
	int * returncode, unsigned int *  jobs,  int jobssize)
{


	double * colB;
	int * copied_counter;
	int * shared_sdiag;
	double * stidA;
	double * stidB;

	__shared__ int * backsubst;
	__shared__ int * rhselim;
	__shared__ int cnt_rhselim;
	__shared__ int cnt_backsubst;
	__shared__ int i1;

	int localidx = 0;
	unsigned int threadstid = 0;
	unsigned int threadstnum = 0;
	unsigned int blockproblemnum = 0;

	bool active = getThreadInfo(data.rhs_threadsperstid, data.rhs_problemsperblock, jobs, jobssize, localidx, threadstid, threadstnum, blockproblemnum);
	active = active && returncode[threadstid]==0;

	if (isAllInactive(active)) return;

//	printf("stid=%d, localidx=%d, threadstnum=%d, blockproblemnum=%d, data.problemsperblock=%d\n", threadstid, localidx, threadstnum, blockproblemnum, data.rhs_problemsperblock);
//	if (threadIdx.x==0)	printf("===================================================\n\nRRSLV\n\n===================================================\n");

	colB = dynamic_shared+blockproblemnum*data.n;
	copied_counter = (int*)(dynamic_shared+data.rhs_problemsperblock*data.n);
	shared_sdiag = copied_counter+data.rhs_problemsperblock;

	exec_arg3(
		m_simple_int_copy, data.n, threadIdx.x, blockDim.x, copied_counter[0], 1, 
		shared_sdiag, data.sdiag, (void*)0);

	stidA = A+threadstid*totalmemsizeA;
	stidB = B+threadstid*totalmemsizeB;

	for (int k=0; k<m; k++)
	{

		__syncthreads();

		exec_arg3(
			m_simple_copy, data.n, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
			colB, stidB+k*colmemsizeB, (void*)0);

		__syncthreads();
		if (threadIdx.x==0)
		{
			rhselim = data.rhselim;
		}

		__syncthreads();

		for (int i=0; i<data.n-1; i++)
		{
			if (threadIdx.x == 0)
			{
				cnt_rhselim = data.n_rhselim[i];
				i1 = shared_sdiag[i]-1;
			}
			__syncthreads();

			double * colAi1 = stidA+i1*colmemsizeA;
	//	printf("error=%p\n", error);

			if (localidx==0 && active)
			{
				colB[i1] /= colAi1[i1];
			}

			__syncthreads();

			exec_arg5(
				m_rhselim, cnt_rhselim, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
				colB, colAi1, rhselim, i1, (void*)0);

			__syncthreads();

			if(threadIdx.x==0)
			{
				rhselim += cnt_rhselim;
			}
		}

		if (threadIdx.x==0)
		{
			backsubst = data.backsubst;
			cnt_backsubst = data.n_backsubst[0];
			i1 = shared_sdiag[data.n-1]-1;
		}
		__syncthreads();

		if (localidx == 0 && active)
		{
			double diag = stidA[i1*colmemsizeA+i1];
			colB[i1] /= (diag*diag);
		}

		__syncthreads();

		for (int i=data.n-1; i>=1; i--)
		{
			exec_arg5(
				m_backsubst, cnt_backsubst, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
				colB, stidA+i1*colmemsizeA, backsubst, i1, (void*)0);

			__syncthreads();

			if (threadIdx.x == 0)
			{
				backsubst += cnt_backsubst;
				cnt_backsubst = data.n_backsubst[data.n-(i-1)-1];
				i1 = shared_sdiag[i-1]-1;
			}

			__syncthreads();

			double diag = stidA[i1*colmemsizeA+i1];

			if (localidx==0 && active)
			{
				colB[i1] /= diag;
			}

			__syncthreads();

		}

		exec_arg3(
			m_simple_copy, data.n, localidx, data.rhs_threadsperstid, copied_counter[blockproblemnum], active, 
			stidB+k*colmemsizeB, colB, (void*)0);

	}

}


void cuda_chlsco(const Scheduler::T_CUDADATA & data, const int & totalmemsizeA, const int & colmemsizeA, double * const & A,
	int * const & returncode, unsigned int * const & jobs, const int & jobssize, cudaStream_t & stream)
{
	int blocks = jobssize/data.problemsperblock;
	if (jobssize%data.problemsperblock!=0) blocks++;
	kernel_chlsco<<<blocks,data.threadsperstid*data.problemsperblock,data.sharedmemsize,stream>>>(data, totalmemsizeA, colmemsizeA, A, returncode, jobs, jobssize);
}

void cuda_chlsco_trans(const Scheduler::T_CUDADATA & data, const int & totalmemsizeA, const int & colmemsizeA, double * const & A,
	int * const & returncode, unsigned int * const & jobs, const int & jobssize, cudaStream_t & stream)
{
	dim3 blockdim(data.trans_rows,data.trans_problemsperblock*data.trans_rows);
	int blocks = jobssize/data.trans_problemsperblock;
	if (jobssize%data.trans_problemsperblock != 0) blocks++;
	kernel_transp<<<blocks,blockdim,data.trans_sharedmemsize,stream>>>(data, totalmemsizeA, colmemsizeA, A, returncode, jobs, jobssize);
}

void cuda_rrslvo(const Scheduler::T_CUDADATA & data,
	const int & totalmemsizeA, const int & colmemsizeA, double * const & A,
	const int & m, const int & totalmemsizeB, const int & colmemsizeB, double * const & B,
	int * const & returncode, unsigned int * const & jobs, const int & jobssize, cudaStream_t & stream)
{
	int blocks = jobssize/data.rhs_problemsperblock;
	if (jobssize%data.rhs_problemsperblock!=0) blocks++;
	kernel_rrslvo<<<blocks,data.rhs_threadsperstid*data.rhs_problemsperblock,data.rhs_sharedmemsize,stream>>>(data,
		totalmemsizeA, colmemsizeA, A,
		m, totalmemsizeB, colmemsizeB, B, 
		returncode, jobs, jobssize);
/*	{
	    cudaError_t cudaerr = cudaDeviceSynchronize();
	    if (cudaerr != cudaSuccess)
	    {
	        printf("kernel_rrslvo: kernel launch failed with error \"%s\".\n",
	               cudaGetErrorString(cudaerr));
	        throw 1;
	    }
	}
*/
//	cudaDeviceSynchronize();
}

	
}

